package otetcd

import (
	"fmt"
	"time"

	"github.com/DoNewsCode/core/config"
	"github.com/DoNewsCode/core/contract"
	"github.com/DoNewsCode/core/di"
	"github.com/DoNewsCode/core/internal"
	"github.com/go-kit/log"
	"github.com/opentracing-contrib/go-grpc"
	"github.com/opentracing/opentracing-go"
	"go.etcd.io/etcd/client/v3"
	"google.golang.org/grpc"
)

/*
Providers returns a set of dependencies including the Maker, the default *clientv3.Client and the exported configs.
	Depends On:
		log.Logger
		contract.ConfigAccessor
		EtcdConfigInterceptor `optional:"true"`
		opentracing.Tracer    `optional:"true"`
	Provide:
		Maker
		Factory
		*clientv3.Client
*/
func Providers() []interface{} {
	return []interface{}{provideFactory, provideDefaultClient, provideConfig}
}

// EtcdConfigInterceptor is an injector type hint that allows user to do
// last minute modification to etcd configurations. This is useful when some
// configuration can not be expressed in yaml/json. For example, the *tls.Config.
type EtcdConfigInterceptor func(name string, options *clientv3.Config)

// Maker is models Factory
type Maker interface {
	Make(name string) (*clientv3.Client, error)
}

// Factory is a *di.Factory that creates *clientv3.Client using a
// specific configuration entry.
type Factory struct {
	*di.Factory
}

// Make creates *clientv3.Client using a specific configuration entry.
func (r Factory) Make(name string) (*clientv3.Client, error) {
	client, err := r.Factory.Make(name)
	if err != nil {
		return nil, err
	}
	return client.(*clientv3.Client), nil
}

// factoryIn is the injection parameter for provideFactory.
type factoryIn struct {
	di.In

	Logger      log.Logger
	Conf        contract.ConfigAccessor
	Interceptor EtcdConfigInterceptor `optional:"true"`
	Tracer      opentracing.Tracer    `optional:"true"`
	Dispatcher  contract.Dispatcher   `optional:"true"`
}

// FactoryOut is the result of Provide.
type FactoryOut struct {
	di.Out

	Maker   Maker
	Factory Factory
}

// provideFactory creates Factory. It is a valid
// dependency for package core.
func provideFactory(p factoryIn) (FactoryOut, func()) {

	factory := di.NewFactory(func(name string) (di.Pair, error) {
		var (
			conf Option
		)
		if err := p.Conf.Unmarshal(fmt.Sprintf("etcd.%s", name), &conf); err != nil {
			return di.Pair{}, fmt.Errorf("etcd configuration %s not valid: %w", name, err)
		}
		if len(conf.Endpoints) == 0 {
			conf.Endpoints = envDefaultEtcdAddrs
		}
		co := clientv3.Config{
			Endpoints:            conf.Endpoints,
			AutoSyncInterval:     duration(conf.AutoSyncInterval),
			DialTimeout:          duration(conf.DialTimeout),
			DialKeepAliveTime:    duration(conf.DialKeepAliveTime),
			DialKeepAliveTimeout: duration(conf.DialKeepAliveTimeout),
			MaxCallSendMsgSize:   conf.MaxCallSendMsgSize,
			MaxCallRecvMsgSize:   conf.MaxCallRecvMsgSize,
			TLS:                  conf.TLS,
			Username:             conf.Username,
			Password:             conf.Password,
			RejectOldCluster:     conf.RejectOldCluster,
			DialOptions:          conf.DialOptions,
			Context:              conf.Context,
			LogConfig:            conf.LogConfig,
			PermitWithoutStream:  conf.PermitWithoutStream,
		}
		if p.Tracer != nil {
			co.DialOptions = append(
				co.DialOptions,
				grpc.WithUnaryInterceptor(otgrpc.OpenTracingClientInterceptor(p.Tracer)),
				grpc.WithStreamInterceptor(otgrpc.OpenTracingStreamClientInterceptor(p.Tracer)),
			)
		}
		if p.Interceptor != nil {
			p.Interceptor(name, &co)
		}
		client, _ := clientv3.New(co)
		return di.Pair{
			Conn: client,
			Closer: func() {
				_ = client.Close()
			},
		}, nil
	})
	etcdFactory := Factory{factory}
	etcdFactory.SubscribeReloadEventFrom(p.Dispatcher)
	out := FactoryOut{
		Maker:   etcdFactory,
		Factory: etcdFactory,
	}
	return out, factory.Close
}

func provideDefaultClient(maker Maker) (*clientv3.Client, error) {
	return maker.Make("default")
}

type configOut struct {
	di.Out

	Config []config.ExportedConfig `group:"config,flatten"`
}

func provideConfig() configOut {
	return configOut{
		Config: []config.ExportedConfig{
			{
				Owner: "otetcd",
				Data: map[string]interface{}{
					"etcd": map[string]Option{
						"default": {
							Endpoints:            envDefaultEtcdAddrs,
							AutoSyncInterval:     config.Duration{},
							DialTimeout:          config.Duration{},
							DialKeepAliveTime:    config.Duration{},
							DialKeepAliveTimeout: config.Duration{},
							MaxCallSendMsgSize:   0,
							MaxCallRecvMsgSize:   0,
							TLS:                  nil,
							Username:             "",
							Password:             "",
							RejectOldCluster:     false,
							DialOptions:          nil,
							Context:              nil,
							LogConfig:            nil,
							PermitWithoutStream:  false,
						},
					},
				},
				Comment: "The configuration for ETCD.",
			},
		},
	}
}

func duration(d config.Duration) time.Duration {
	return d.Duration
}

var envDefaultEtcdAddrs, envDefaultEtcdAddrsIsSet = internal.GetDefaultAddrsFromEnv("ETCD_ADDR", "127.0.0.1:2379")
